/*****************************************************************************/
/*
                                  cookie.c

COPYRIGHT
---------
Copyright (C) 2008-2013 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY
---------------
24-MAR-2008  MGD  initial
*/

/*****************************************************************************/

#ifdef VWCMS_VMS_V62
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60200000
#undef __CRTL_VER
#define __CRTL_VER 60200000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* VMS related header files */
#include "ssdef.h"
#include "stsdef.h"

/* application header file */
#include "cgilib.h"
#include "config.h"
#include "content.h"
#include "cookie.h"
#include "other.h"
#include "vwcms.h"

#define FI_LI "COOKIE", __LINE__

/* global storage */

static int CookieSetLength;

static char *CookieSetPtr;

/* prototypes */

int sys$fao(__unknown_params);

/* external storage */

extern BOOL  Debug,
             WatchEnabled;

extern char  ContentSiteName[],
             SoftwareId[],
             Utility[];

extern CONFIG_DATA  VWcmsConfig;

/*****************************************************************************/
/*
Returns a dynamically allocated string containing the cookie value, or NULL.
*/

char* CookieGet (char *CookieName)

{
   char  *cptr, *sptr,
         *CookieValuePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CookieGet() |%s|\n", CookieName);

   cptr = GetCgiVarNull ("HTTP_COOKIE");
   if (!cptr)
   {
      if (WatchEnabled) WatchThis (FI_LI, "!AZ=(null)", CookieName);
      return (NULL);
   }

   cptr = strstr (cptr, CookieName);
   if (!cptr) return (NULL);
   cptr += strlen(CookieName);
   if (*cptr != '=') return (NULL);
   cptr++;
   for (sptr = cptr; *sptr && *sptr != ';' && *sptr != ' '; sptr++);

   CookieValuePtr = CgiLibVeeMemCalloc (sptr-cptr);
   if (!CookieValuePtr) ErrorExit (vaxc$errno, FI_LI);
   memcpy (CookieValuePtr, cptr, sptr-cptr);

   if (WatchEnabled) WatchThis (FI_LI, "!AZ=!AZ", CookieName, CookieValuePtr);

   return (CookieValuePtr);
}

/*****************************************************************************/
/*
Create a field suitable for inclusion in an HTTP response to set the supplied
cookie value.  If 'CookieValue' is NULL then resent the 'CookieName'. 
*/ 

void CookieSet
(
char *CookieName,
char *CookieValue
)
{
   static $DESCRIPTOR (SetCookieFaoDsc,
                       "Set-Cookie: !AZ=!AZ; path=!AZ; domain=!AZ;!AZ\n");
   static char PreExpired [] = " expires=Fri, 13 Jan 1978 14:00:00 GMT;";

   int  status;
   unsigned short  slen;
   char  *cptr, *sptr, *zptr,
         *PathPtr,
         *SetCookiePtr;
   char  DomainBuffer [128],
         SetCookieBuffer [1024];
   $DESCRIPTOR (SetCookieBufferDsc, SetCookieBuffer);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "CookieSet() |%s|%s|\n", CookieName, CookieValue);

   if (!CookieName)
   {
      /* (re)initialise */
      CookieSetPtr = NULL;
      CookieSetLength = 0;
      return;
   }

   cptr = GetCgiVarNull ("HTTP_HOST");
   if (!cptr || !*cptr) cptr = GetCgiVar("SERVER_NAME");
   if (!cptr || !*cptr) ErrorExit (SS$_BUGCHECK, FI_LI);
   /* need to remove any potential non-well-known :port */
   zptr = (sptr = DomainBuffer) + sizeof(DomainBuffer)-1;
   while (*cptr && *cptr != ':' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';

   if (VWcmsConfig.SiteRoot)
      PathPtr = "/";
   else
      PathPtr = ContentEncodeName ("/",NULL);

   status = sys$fao (&SetCookieFaoDsc, &slen, &SetCookieBufferDsc,
                     CookieName,
                     CookieValue ? CookieValue : "",
                     PathPtr,
                     DomainBuffer,
                     CookieValue ? "" : PreExpired);
   if (VMSnok(status)) ErrorExit (status, FI_LI);

   CookieSetPtr = CgiLibVeeMemRealloc (CookieSetPtr,
                                       CookieSetLength + slen + 2);
   if (!CookieSetPtr) ErrorExit (vaxc$errno, FI_LI);
   memcpy (CookieSetPtr+CookieSetLength, SetCookieBuffer, slen);
   CookieSetLength += slen;

   if (Debug) fprintf (stdout, "%d |%s|\n", CookieSetLength, CookieSetPtr);
}

/*****************************************************************************/
/*
Return a pointer to a "Set-Cookie: ..." HTTP header if one has been set or to
an empty string otherwise.  Reset the global pointer as necessary.
*/

char* CookieHeader ()

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CookieHeader()\n");

   if (cptr = CookieSetPtr)
      CookieSetPtr = NULL;
   else
      cptr = "";

   return (cptr);
}

/*****************************************************************************/

