/*****************************************************************************/
/*
                                  ticket.c

(Perhaps lame) attempt to frustrate SPAMbots.

A 'ticket' is a token that is searched for in submitted text and gives
permission for a comment/feedback/form submission to be accepted.

The ticket is a 15 character string of base-64 encoded time components
providing a pseudo-random collection of printable characters.

Issued tickets are stored in a site-specific file and remain valid for a
fixed period.  This file is opened and every (still-)valid ticket is searched
for in the supplied text.  If hit the text has been verified and the ticket
immediately invalidated.


COPYRIGHT
---------
Copyright (C) 2008-2013 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY
---------------
06-AUG-2011  MGD  TicketPresent() if not found in content look in template
30-JUL-2011  MGD  bugfix; TicketFile()
02-DEC-2009  MGD  initial
*/

/*****************************************************************************/

#ifdef VWCMS_VMS_V62
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60200000
#undef __CRTL_VER
#define __CRTL_VER 60200000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <locale.h>
#include <stat.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <types.h>
#include <time.h>
#include <unistd.h>

/* VMS related header files */
#include "descrip.h"
#include "rmsdef.h"
#include "ssdef.h"
#include "stsdef.h"

/* application header file */
#include "cgilib.h"
#include "config.h"
#include "content.h"
#include "data.h"
#include "form.h"
#include "other.h"
#include "template.h"
#include "ticket.h"
#include "vwcms.h"

#define FI_LI "TICKET", __LINE__

/* prototypes */

/* global storage */

BOOL  TicketOnPage;

static char  *TicketTestPtr;

/* external storage */

extern BOOL  Debug,
             WatchEnabled;

extern int  CurrentSecondsTime;

extern unsigned long  CurrentVmsBinTime[];
extern unsigned short  CurrentVmsNumTime[];

extern char  SoftwareId[],
             Utility[];

extern char  *TemplateTextPtr;

extern CONTENT_DATA  *ContentSectionPtr;

extern CONFIG_DATA  VWcmsConfig;

/*****************************************************************************/
/*
Issue a comment/feedback/form ticket.  The directive parameter may have
multiple '$'-delimitted prompts describing the inclusion of the ticket into
the text.  These are rotated on a pseudo-random basis as an additional
obfuscation factor against a 'bot.  Useful?  Who knows!
*/

void TicketIssue (char* SubsArg)

{
   BOOL ExampleOnly = FALSE;
   int  status,
        PhraseCount;
   unsigned short TicketVal [6];
   char  *cptr, *sptr, *zptr;
   char  ArgString [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TicketIssue() |%s|\n", SubsArg);

   if (!SubsArg)
   {
      /* initialisation */
      TicketOnPage = FALSE;
      return;
   }

   if (*SubsArg == '$') SubsArg++;
   if (*SubsArg == '~')
   {
      SubsArg++;
      ExampleOnly = TRUE;
   }

   PhraseCount = 1;
   for (cptr = SubsArg; *cptr; cptr++)
      if (*cptr == '$') PhraseCount++;
   cptr = SubsArg;
   PhraseCount = CurrentVmsNumTime[6] % PhraseCount;
   while (PhraseCount)
   {
      while (*cptr && *cptr != '$') cptr++;
      if (*cptr) cptr++;
      PhraseCount--;
   }
   zptr = (sptr = ArgString) + sizeof(ArgString)-1;
   while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';

   TicketVal[0] = ((USHORTPTR)CurrentVmsBinTime)[0];
   TicketVal[1] = ((USHORTPTR)CurrentVmsBinTime)[2];
   TicketVal[2] = TicketVal[0] ^ TicketVal[1];
   TicketVal[3] = ((USHORTPTR)CurrentVmsBinTime)[1];
   TicketVal[4] = TicketVal[1] ^ TicketVal[3];
   TicketVal[5] = TicketVal[0] ^ ((USHORTPTR)CurrentVmsBinTime)[3];

   /* base-64 encode of 12 chars produces 16 chars */
   cptr = Base64Encode ((char*)TicketVal, sizeof(TicketVal));
   /* chop that back to 15 for ease of handling from TicketFile() */
   cptr[TICKET_SIZE] = '\0';

   if (ExampleOnly)
   {
      status = SS$_NORMAL;
      TicketTestPtr = cptr;
   }
   else
   {
      TicketTestPtr = NULL;
      status = TicketFile (cptr, NULL);
   }

   if (VMSok (status))
   {
      if (!ArgString[0] || ArgString[0] == '#')
      {
         /* just a little more obfuscation */
         static char  ColHex [] = { 'a','b','c','d','e' };
         static char  *TagArray [] = { "span", "code", "var", "span",
                                       "span", "kbd", "cite", "span", };
         static char  *StyleArray [] = {
"font-style: normal; background-color: %s; padding: 0 11px 0 11px;",
"background-color: %s; font-style: normal; padding: 0 10px 0 10px;",
"padding: 0 9px 0 9px; background-color: %s; font-style: normal;",
"background-color: %s; padding: 0 10px 0 10px; font-style: normal;"
};

         if (!ArgString[0])
            sprintf (ArgString, "#%c%c%c%c%c%c",
                     ColHex[(CurrentVmsNumTime[4]%5)],
                     ColHex[(CurrentVmsNumTime[5]%5)],
                     ColHex[(CurrentVmsNumTime[6]%5)],
                     ColHex[(CurrentVmsNumTime[4]%5)],
                     ColHex[(CurrentVmsNumTime[5]%5)],
                     ColHex[(CurrentVmsNumTime[6]%5)]);

         fprintf (stdout, "<%s style=\"",
                  TagArray[(CurrentVmsNumTime[6]%8)]);
         fprintf (stdout, StyleArray[(CurrentVmsNumTime[5]%4)], ArgString);
         fprintf (stdout, "\">%s</%s>",
                  cptr, TagArray[(CurrentVmsNumTime[6]%8)]);
      }
      else
         fprintf (stdout, "%s %s", ArgString, cptr);

      TicketOnPage = TRUE;
   }
   else
      TemplateAdvise ("");
}

/*****************************************************************************/
/*
The ticket is a 15 character string (base-64 encoded time components).  It is
stored as 15 character plus 1 null string (really just for strstr()
convenience).  The ticket is preceded by a four byte integer (the U*x time in
seconds).  This makes each entry 20 bytes.

This function operates in two modes depending on which parameter has been
supplied.

1) Issue ticket; a $$TICKET$$ directive has been processed and a ticket string
displayed on the page.  This string is then stored in the ticket file
associated with the site along with it's time of issue.

2) Ticket text; a comment or feedback (email) page has submitted the text for
processing (comment registration or email transmission).  Each ticket in the
file is in turn searched for in the text (using strstr()).  If hit the
submission contained a legitimate ticket.  The ticket string is cut from the
text.

During file scan (either mode) ticket timestamps are checked for expiry and
invalidated as required.

The function returns a VMS success status if successful or an error status if
not.  VWcms advice is also issue as appropriate.
*/

int TicketFile
(
char *IssueTicket,
char *TicketText
)
{
   BOOL  HitTicket;
   int  status,
        ContentLength,
        ValidCount;
   char  *cptr, *sptr, *zptr,
         *ContentPtr,
         *FileNamePtr,
         *SlotPtr;
   FILE  *FilePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TicketFile()\n");

   if (ConfigReadOnly (TRUE)) return (SS$_ABORT);

   if (IssueTicket && strlen(IssueTicket) != TICKET_SIZE)
   {
      status = SS$_BUGCHECK;
      TemplateAdvise ("ECOMMENT !AZ (%X!8XL)", SysGetMsg(status), status);
      return (status);
   }

   FileNamePtr = MakeFileName ("TICKET", ".DAT");

   status = WaitForLock (FileNamePtr, TRUE);
   if (VMSok (status))
      status = DataReadFileIntoMemory (FileNamePtr, &ContentPtr,
                                       &ContentLength, FALSE);
   if (VMSnok (status))
   {
      if (status != RMS$_FNF)
      {
         WaitForUnLock ();
         TemplateAdvise ("ECOMMENT !AZ (%X!8XL)", SysGetMsg(status), status);
         return (status);
      }
      /* file does not yet exist */
      ContentLength = 0;
      ContentPtr = CgiLibVeeMemCalloc (ContentLength + 32);
   }

   SlotPtr = NULL;
   HitTicket = FALSE;
   ValidCount = 0;

   /* scan through each 20 byte entry */
   zptr = (cptr = ContentPtr) + ContentLength;
   while (cptr < zptr)
   {
      if (*(unsigned int*)cptr == 0xffff)
      {
         /* note the first reusable slot */
         if (!SlotPtr) SlotPtr = cptr;
      }
      else
      {
         if (*(unsigned int*)cptr + COMMENT_TICKET_TIMEOUT < CurrentSecondsTime)
         {
            /* timeout - invalidate the timestamp */
            *(unsigned int*)cptr = 0xffff;
            /* note the first reusable slot */
            if (!SlotPtr) SlotPtr = cptr;
         }
         else
         if (IssueTicket)
         {
            /* check if the supplied ticket is already in the file */
            if (MATCH0(cptr+4,(void*)IssueTicket,16))
            {
               HitTicket = TRUE;
               break;
            }
            ValidCount++;
         }
         else
         {
            /* look for this ticket in the supplied text */
            if (sptr = strstr (TicketText, cptr+4))
            {
               /* once-only ticket - invalidate it now */
               *(unsigned int*)cptr = 0xffff;
               HitTicket = TRUE;
               break;
            }
            ValidCount++;
         }
      }
      cptr += sizeof(int)+16;
   }

   /* if file is attempting to grow beyond limits */
   if (ValidCount >= COMMENT_TICKET_VALID_MAX)
   {
      /* bogus message but there we go! */
      status = SS$_REGISFULL;
      WaitForUnLock ();
      TemplateAdvise ("ECOMMENT !AZ (%X!8XL)", SysGetMsg(status), status);
      return (status);
   }

   if (IssueTicket)
   {
      if (HitTicket)
      {
         /* found the same ticket! */
         status = SS$_BUGCHECK;
         WaitForUnLock ();
         TemplateAdvise ("ECOMMENT !AZ (%X!8XL)", SysGetMsg(status), status);
         return (status);
      }
      if (!SlotPtr)
      {
         /* append to current tickets */
         SlotPtr = ContentPtr + ContentLength;
         /* DataReadFileIntoMemory() allocates an extra 32 bytes! */
         ContentLength += 4 + 16;
      }
      /* put the ticket into the slot ready to write */
      *(unsigned int*)SlotPtr = CurrentSecondsTime;
      memcpy (SlotPtr+4, (void*)IssueTicket, 16);
      status = SS$_NORMAL;
   }
   else
   {
      if (HitTicket)
      {
         TicketExcise (TicketText, sptr);
         status = SS$_NORMAL;
      }
      else
      {
         TemplateAdvise ("W!AZ", CONFIG_LANG_PARSE(LangModerate,2));
         status = SS$_ABORT;
      }
   }

   FilePtr = fopen (FileNamePtr, "w+", "ctx=bin",
                    "rfm=stmlf", "shr=nil", "fop=cif");
   if (!FilePtr)
   {
      status = vaxc$errno;
      WaitForUnLock ();
      TemplateAdvise ("ECOMMENT !AZ (%X!8XL)", SysGetMsg(status), status);
      return (status);
   }

   rewind (FilePtr);
   fwrite (ContentPtr, ContentLength, 1, FilePtr);
   fclose (FilePtr);

   if (chmod (FileNamePtr, 0700)) ErrorExit (vaxc$errno, FI_LI);

   WaitForUnLock ();

   return (status);
}

/*****************************************************************************/
/*
Remove the ticket from the text.
Expect it be a separate line though it doesn't have to be.
*/

void TicketExcise
(
char *TextPtr,
char *TicketPtr
)
{
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TicketExcise() |%s|\n", TextPtr);

   sptr = (cptr = TicketPtr) + TICKET_SIZE;
   while (cptr > TextPtr && isspace(*(cptr-1)) && *(cptr-1) != '\n') cptr--;
   while (*sptr && isspace(*sptr) && *sptr != '\n') sptr++;
   if (*sptr == '\n') sptr++;
   while (*sptr) *cptr++ = *sptr++;
   *cptr = '\0';

   if (Debug) fprintf (stdout, "|%s|\n", TextPtr);
}

/*****************************************************************************/
/*
Return the ticket 'test' pointer (or reset it if called with TRUE).
*/

char* TicketTest (BOOL ResetTest)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TicketTest()\n");

   if (ResetTest) TicketTestPtr = NULL;
   return (TicketTestPtr);
}

/*****************************************************************************/
/*
Returns a pointer if the section content or template contains a $$TICKET$$
directive, else NULL.
*/

char* TicketPresent ()

{
   char  *cptr;
   char  **taptr;
   char  *TextArray[] = { ContentSectionPtr->TextPtr, TemplateTextPtr, NULL };

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TicketPresent()\n");

   for (taptr = TextArray; *taptr; taptr++)
      if (*(cptr = TemplateParseFind (*taptr, "$$TICKET$")))
         return (cptr);

   return (NULL);
}

/*****************************************************************************/

